@@ -23,7 +23,9 @@
 import contextlib
 import functools
 import numpy as np
+import statistics
 
+from collections import Counter
 from dataiter import aggregate
 from dataiter import util
 
@@ -43,7 +45,7 @@ except Exception:
 with contextlib.suppress(LookupError):
     USE_NUMBA = util.parse_env_boolean("DATAITER_USE_NUMBA")
 
-__version__ = "0.29"
+__version__ = "0.29.1"
 
 DEFAULT_PEEK_ELEMENTS = 10
 DEFAULT_PEEK_ITEMS = 3
@@ -145,7 +147,7 @@ def count(x="", dropna=False):
             f.fallback = fallback
             return f
         return fallback
-    if dropna:
+    if x.is_float() and dropna:
         x = x[~np.isnan(x)]
     return len(x)
 
@@ -168,9 +170,9 @@ def count_unique(x, dropna=False):
             f.fallback = fallback
             return f
         return fallback
-    if dropna:
+    if x.is_float() and dropna:
         x = x[~np.isnan(x)]
-    return len(np.unique(x))
+    return len(set(x))
 
 @_ensure_x_type
 def first(x):
@@ -217,7 +219,7 @@ def max(x, dropna=True):
             f.fallback = fallback
             return f
         return fallback
-    if dropna:
+    if x.is_float() and dropna:
         x = x[~np.isnan(x)]
     if len(x) < 1:
         return np.nan
@@ -245,7 +247,7 @@ def mean(x, dropna=True):
             f.fallback = fallback
             return f
         return fallback
-    if dropna:
+    if x.is_float() and dropna:
         x = x[~np.isnan(x)]
     if len(x) < 1:
         return np.nan
@@ -273,7 +275,7 @@ def median(x, dropna=True):
             f.fallback = fallback
             return f
         return fallback
-    if dropna:
+    if x.is_float() and dropna:
         x = x[~np.isnan(x)]
     if len(x) < 1:
         return np.nan
@@ -298,7 +300,7 @@ def min(x, dropna=True):
             f.fallback = fallback
             return f
         return fallback
-    if dropna:
+    if x.is_float() and dropna:
         x = x[~np.isnan(x)]
     if len(x) < 1:
         return np.nan
@@ -323,12 +325,18 @@ def mode(x, dropna=True):
             f.fallback = fallback
             return f
         return fallback
-    if dropna:
+    if x.is_float() and dropna:
         x = x[~np.isnan(x)]
     if len(x) < 1:
         return x.missing_value
-    values, counts = np.unique(x, return_counts=True)
-    return values[counts.argmax()]
+    # We could use np.unique here, but it makes getting
+    # the first one in case of ties really complicated.
+    try:
+        return statistics.mode(x)
+    except statistics.StatisticsError:
+        # Python < 3.8 with several elements tied for mode,
+        # will return one of the tied elements at random.
+        return Counter(x).most_common(1)[0][0]
 
 def nrow(data):
     """
@@ -394,7 +402,7 @@ def quantile(x, q, dropna=True):
             f.fallback = fallback
             return f
         return fallback
-    if dropna:
+    if x.is_float() and dropna:
         x = x[~np.isnan(x)]
     if len(x) < 1:
         return np.nan
@@ -431,7 +439,7 @@ read_json.__doc__ = util.format_alias_doc(read_json, ListOfDicts.read_json)
 read_npz.__doc__ = util.format_alias_doc(read_npz, DataFrame.read_npz)
 
 @_ensure_x_type
-def std(x, dropna=True):
+def std(x, ddof=0, dropna=True):
     """
     Return the standard deviation of `x`.
 
@@ -446,17 +454,19 @@ def std(x, dropna=True):
     """
     if isinstance(x, str):
         def fallback(data):
-            return std(data[x], dropna=dropna)
-        if USE_NUMBA:
+            return std(data[x], ddof=ddof, dropna=dropna)
+        # Numba doesn't support the ddof argument,
+        # so can only handle the default ddof=0.
+        if USE_NUMBA and ddof == 0:
             f = aggregate.generic(x, np.std, dropna=dropna, default=np.nan, nrequired=2)
             f.fallback = fallback
             return f
         return fallback
-    if dropna:
+    if x.is_float() and dropna:
         x = x[~np.isnan(x)]
     if len(x) < 2:
         return np.nan
-    return np.std(x).item()
+    return np.std(x, ddof=ddof).item()
 
 @_ensure_x_type
 def sum(x, dropna=True):
@@ -477,12 +487,12 @@ def sum(x, dropna=True):
             f.fallback = fallback
             return f
         return fallback
-    if dropna:
+    if x.is_float() and dropna:
         x = x[~np.isnan(x)]
     return np.sum(x).item()
 
 @_ensure_x_type
-def var(x, dropna=True):
+def var(x, ddof=0, dropna=True):
     """
     Return the variance of `x`.
 
@@ -497,14 +507,16 @@ def var(x, dropna=True):
     """
     if isinstance(x, str):
         def fallback(data):
-            return var(data[x], dropna=dropna)
-        if USE_NUMBA:
+            return var(data[x], ddof=ddof, dropna=dropna)
+        # Numba doesn't support the ddof argument,
+        # so can only handle the default ddof=0.
+        if USE_NUMBA and ddof == 0:
             f = aggregate.generic(x, np.var, dropna=dropna, default=np.nan, nrequired=2)
             f.fallback = fallback
             return f
         return fallback
-    if dropna:
+    if x.is_float() and dropna:
         x = x[~np.isnan(x)]
     if len(x) < 2:
         return np.nan
-    return np.var(x).item()
+    return np.var(x, ddof=ddof).item()
