@@ -20,13 +20,13 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
 
+import contextlib
 import dataiter
 import datetime
 import itertools
 import json
 import numpy as np
 import pickle
-import sys
 
 from dataiter import deco
 from dataiter import util
@@ -38,7 +38,7 @@ class DataFrameColumn(Vector):
     """
     A column in a data frame.
 
-    DataFrameColumn is a subclass of :class:`Vector`. See the vector
+    DataFrameColumn is a subclass of :class:`.Vector`. See the vector
     documentation for relevant properties and methods.
     """
 
@@ -95,7 +95,7 @@ class DataFrame(dict):
     A class for tabular data.
 
     DataFrame is a subclass of ``dict``, with columns being
-    :class:`DataFrameColumn`, which are :class:`Vector`, which are NumPy
+    :class:`.DataFrameColumn`, which are :class:`.Vector`, which are NumPy
     ``ndarray``. This means that basic ``dict`` methods, such as ``items()``,
     ``keys()`` and ``values()`` can be used iterate over and manage the data as
     a whole and NumPy functions and array methods can be used for fast
@@ -157,7 +157,7 @@ class DataFrame(dict):
     def __getattr__(self, name):
         if name in self:
             return self.__getitem__(name)
-        return super().__getattr__(name)
+        raise AttributeError(name)
 
     def __setattr__(self, name, value):
         if name in self.ATTRIBUTES:
@@ -183,21 +183,41 @@ class DataFrame(dict):
 
         In `colname_function_pairs`, `function` receives as an argument a data
         frame object, a group-wise subset of all rows. It should return a
-        scalar value.
+        scalar value. Common aggregation functions have shorthand helpers
+        available under :mod:`dataiter`, see the guide on :doc:`aggregation
+        </aggregation>` for details.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
-        >>> data.group_by("hood").aggregate(n=di.nrow, price=lambda x: np.nanmean(x.price))
+        >>> data = di.read_csv("data/listings.csv")
+        >>> # The below aggregations are identical. Usually you'll get by
+        >>> # with the shorthand helpers, but for complicated calculations,
+        >>> # you might need custom lambda functions.
+        >>> data.group_by("hood").aggregate(n=di.count(), price=di.mean("price"))
+        >>> data.group_by("hood").aggregate(n=lambda x: x.nrow, price=lambda x: np.nanmean(x.price))
         """
-        by = np.column_stack([self[x].as_bytes() for x in self._group_colnames])
-        values, ui, inv = np.unique(by, return_index=True, return_inverse=True, axis=0)
-        stat = self.slice(ui).select(*self._group_colnames)
-        slice_indices = {}
-        for i in range(self.nrow):
-            slice_indices.setdefault(inv[i], []).append(i)
-        slices = [self.slice(slice_indices[i]) for i in range(stat.nrow)]
+        group_colnames = self._group_colnames
+        data = self.sort(**dict.fromkeys(group_colnames, 1))
+        data._index_ = np.arange(data.nrow)
+        stat = data.unique(*group_colnames).select("_index_", *group_colnames)
+        indices = np.split(data._index_, stat._index_[1:])
+        uses_numba = [getattr(x, "numba", False) for x in colname_function_pairs.values()]
+        if any(uses_numba):
+            groups = Vector.fast(range(len(indices)), int)
+            n = Vector.fast(map(len, indices), int)
+            data._group_ = np.repeat(groups, n)
+        slices = None
         for colname, function in colname_function_pairs.items():
+            uses_numba = getattr(function, "numba", False)
+            if uses_numba:
+                # Numba-accelerated aggregation functions raise
+                # NotImplementedError if given an unsupported dtype.
+                with contextlib.suppress(NotImplementedError):
+                    stat[colname] = function(data)
+                    continue
+                function = function.fallback
+            if slices is None:
+                slices = [data._view_rows(x) for x in indices]
             stat[colname] = [function(x) for x in slices]
-        return stat.sort(**dict.fromkeys(self._group_colnames, 1))
+        return stat.unselect("_index_", "_group_")
 
     @deco.new_from_generator
     def anti_join(self, other, *by):
@@ -209,8 +229,8 @@ class DataFrame(dict):
         `other`.
 
         >>> # All listings that don't have reviews
-        >>> listings = di.DataFrame.read_csv("data/listings.csv")
-        >>> reviews = di.DataFrame.read_csv("data/listings-reviews.csv")
+        >>> listings = di.read_csv("data/listings.csv")
+        >>> reviews = di.read_csv("data/listings-reviews.csv")
         >>> listings.anti_join(reviews, "id")
         """
         by1, by2 = self._split_by(*by)
@@ -224,7 +244,7 @@ class DataFrame(dict):
         """
         Return data frame with columns from `others` added.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.cbind(di.DataFrame(x=1))
         """
         found_colnames = set()
@@ -247,7 +267,7 @@ class DataFrame(dict):
         """
         Get or set column names as a list.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.head()
         >>> data.colnames
         >>> data.colnames = ["a", "b", "c", "d", "e", "f"]
@@ -292,7 +312,7 @@ class DataFrame(dict):
         `colname_value_pairs` provides a shorthand to check against a fixed
         value. See the example below of equivalent filtering all three ways.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.filter((data.hood == "Manhattan") & (data.guests == 2))
         >>> data.filter(lambda x: (x.hood == "Manhattan") & (x.guests == 2))
         >>> data.filter(hood="Manhattan", guests=2)
@@ -321,7 +341,7 @@ class DataFrame(dict):
         `colname_value_pairs` provides a shorthand to check against a fixed
         value. See the example below of equivalent filtering all three ways.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.filter_out(data.hood == "Manhattan")
         >>> data.filter_out(lambda x: x.hood == "Manhattan")
         >>> data.filter_out(hood="Manhattan")
@@ -338,25 +358,38 @@ class DataFrame(dict):
             yield colname, np.delete(column, rows)
 
     @classmethod
-    def from_json(cls, string, **kwargs):
+    def from_json(cls, string, columns=[], dtypes={}, **kwargs):
         """
         Return a new data frame from JSON `string`.
+
+        `columns` is an optional list of columns to limit to. `dtypes` is an
+        optional dict mapping column names to NumPy datatypes. `kwargs` are
+        passed to ``json.load``.
         """
-        obj = json.loads(string, **kwargs)
-        if not isinstance(obj, list):
+        data = json.loads(string, **kwargs)
+        if not isinstance(data, list):
             raise TypeError("Not a list")
-        keys = util.unique_keys(itertools.chain(*obj))
-        columns = {k: [x.get(k, None) for x in obj] for k in keys}
-        return cls(**columns)
+        keys = util.unique_keys(itertools.chain(*data))
+        if columns:
+            keys = [x for x in keys if x in columns]
+        data = {k: [x.get(k, None) for x in data] for k in keys}
+        for name, dtype in dtypes.items():
+            data[name] = DataFrameColumn(data[name], dtype)
+        return cls(**data)
 
     @classmethod
-    def from_pandas(cls, data):
+    def from_pandas(cls, data, dtypes={}):
         """
         Return a new data frame from ``pandas.DataFrame`` `data`.
+
+        `dtypes` is an optional dict mapping column names to NumPy datatypes.
         """
         # It would be a lot faster to skip tolist here,
         # but then we'd need to map some Pandas dtype oddities.
-        return cls(**{x: data[x].to_numpy().tolist() for x in data.columns})
+        data = {x: data[x].to_numpy().tolist() for x in data.columns}
+        for name, dtype in dtypes.items():
+            data[name] = DataFrameColumn(data[name], dtype)
+        return cls(**data)
 
     def full_join(self, other, *by):
         """
@@ -370,14 +403,20 @@ class DataFrame(dict):
         column names if the correspoding column name differs between `self` and
         `other`.
 
-        >>> listings = di.DataFrame.read_csv("data/listings.csv")
-        >>> reviews = di.DataFrame.read_csv("data/listings-reviews.csv")
+        >>> listings = di.read_csv("data/listings.csv")
+        >>> reviews = di.read_csv("data/listings-reviews.csv")
         >>> listings.full_join(reviews, "id")
         """
-        other = other.modify(_id_=np.arange(other.nrow))
-        a = self.left_join(other, *by)
-        b = other.filter_out(np.isin(other._id_, a._id_))
-        return a.rbind(b).unselect("_id_")
+        a = self.modify(_aid_=np.arange(self.nrow))
+        b = other.modify(_bid_=np.arange(other.nrow))
+        ab = a.left_join(b, *by)
+        # Check which rows of b were not joined into a,
+        # if no rows remain, full join is the same as left join ab.
+        b = b.anti_join(ab, "_bid_")
+        if b.nrow == 0:
+            return ab.unselect("_aid_", "_bid_")
+        ba = b.left_join(a, *by)
+        return ab.rbind(ba).sort(_aid_=1, _bid_=1).unselect("_aid_", "_bid_")
 
     def _get_join_indices(self, other, by1, by2):
         other_ids = list(zip(*[other[x] for x in by2]))
@@ -399,7 +438,7 @@ class DataFrame(dict):
         """
         Return the first `n` rows.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.head(5)
         """
         if n is None:
@@ -420,8 +459,8 @@ class DataFrame(dict):
         column names if the correspoding column name differs between `self` and
         `other`.
 
-        >>> listings = di.DataFrame.read_csv("data/listings.csv")
-        >>> reviews = di.DataFrame.read_csv("data/listings-reviews.csv")
+        >>> listings = di.read_csv("data/listings.csv")
+        >>> reviews = di.read_csv("data/listings-reviews.csv")
         >>> listings.inner_join(reviews, "id")
         """
         by1, by2 = self._split_by(*by)
@@ -447,8 +486,8 @@ class DataFrame(dict):
         column names if the correspoding column name differs between `self` and
         `other`.
 
-        >>> listings = di.DataFrame.read_csv("data/listings.csv")
-        >>> reviews = di.DataFrame.read_csv("data/listings-reviews.csv")
+        >>> listings = di.read_csv("data/listings.csv")
+        >>> reviews = di.read_csv("data/listings-reviews.csv")
         >>> listings.left_join(reviews, "id")
         """
         by1, by2 = self._split_by(*by)
@@ -465,6 +504,23 @@ class DataFrame(dict):
             new[found] = column[src[found]]
             yield colname, new.copy()
 
+    def map(self, function):
+        """
+        Apply `function` to each row in data.
+
+        `function` receives as arguments the full data frame and the loop
+        index. The return value will be a list of whatever `function` returns.
+
+        Note that `map` is an inefficient method as it iterates over rows
+        instead of doing vectorized computation. `map` is mostly intended for
+        complicated conditional cases that are difficult to express in
+        vectorized form.
+
+        >>> data = di.read_csv("data/listings-reviews.csv")
+        >>> data.map(lambda x, i: min(1, x.reviews[i]/100) * min(1, x.rating[i]/100))
+        """
+        return [function(self, i) for i in range(self.nrow)]
+
     @deco.new_from_generator
     def modify(self, **colname_value_pairs):
         """
@@ -479,7 +535,7 @@ class DataFrame(dict):
         data.guests``. `modify` just allows you to do the same in a method
         chain context.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.modify(price_per_guest=data.price/data.guests)
         >>> data.modify(price_per_guest=lambda x: x.price / x.guests)
         """
@@ -494,7 +550,7 @@ class DataFrame(dict):
         """
         Return the amount of columns.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.ncol
         """
         self._check_dimensions()
@@ -508,7 +564,7 @@ class DataFrame(dict):
         """
         Return the amount of rows.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.nrow
         """
         if not self: return 0
@@ -541,7 +597,7 @@ class DataFrame(dict):
         but since it's a method, you can use it at the end of a method chain
         instead of wrapping a ``print`` call around the whole chain.
 
-        >>> di.DataFrame.read_csv("data/listings.csv").print_()
+        >>> di.read_csv("data/listings.csv").print_()
         """
         print(self.to_string(max_rows, max_width))
 
@@ -549,7 +605,7 @@ class DataFrame(dict):
         """
         Print memory use by column and total.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.print_memory_use()
         """
         mem = DataFrame()
@@ -557,10 +613,7 @@ class DataFrame(dict):
             new = DataFrame(column=name)
             new.dtype = str(column.dtype)
             new.item_size = column.itemsize
-            if column.is_object():
-                new.total_size = sum(sys.getsizeof(x) for x in column)
-            else:
-                new.total_size = column.nbytes
+            new.total_size = column.get_memory_use()
             mem = mem.rbind(new)
         new = DataFrame(column="TOTAL")
         new.dtype = "--"
@@ -577,7 +630,7 @@ class DataFrame(dict):
         """
         Print counts of missing values by column.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.print_missing_counts()
         """
         nas = DataFrame()
@@ -595,7 +648,7 @@ class DataFrame(dict):
         """
         Return data frame with rows from `others` added.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.rbind(data)
         """
         data_frames = [self] + list(others)
@@ -614,36 +667,62 @@ class DataFrame(dict):
             yield colname, total
 
     @classmethod
-    def read_csv(cls, fname, encoding="utf_8", header=True, columns=None, sep=","):
+    def read_csv(cls, path, encoding="utf-8", sep=",", header=True, columns=[], dtypes={}):
         """
-        Return a new data frame from CSV file `fname`.
+        Return a new data frame from CSV file `path`.
+
+        Will automatically decompress if `path` ends in ``.bz2|.gz|.xz``.
+
+        `columns` is an optional list of columns to limit to. `dtypes` is an
+        optional dict mapping column names to NumPy datatypes.
         """
         import pandas as pd
-        data = pd.read_csv(fname,
+        data = pd.read_csv(path,
                            sep=sep,
                            header=0 if header else None,
-                           usecols=columns,
+                           usecols=columns or None,
+                           dtype=dtypes,
                            parse_dates=False,
-                           encoding=encoding)
+                           encoding=encoding,
+                           low_memory=False)
 
         if not header:
             data.columns = util.generate_colnames(len(data.columns))
-        return cls.from_pandas(data)
+        return cls.from_pandas(data, dtypes)
 
     @classmethod
-    def read_json(cls, fname, encoding="utf_8", **kwargs):
+    def read_json(cls, path, encoding="utf-8", columns=[], dtypes={}, **kwargs):
         """
-        Return a new data frame from JSON file `fname`.
+        Return a new data frame from JSON file `path`.
+
+        Will automatically decompress if `path` ends in ``.bz2|.gz|.xz``.
+
+        `columns` is an optional list of columns to limit to. `dtypes` is an
+        optional dict mapping column names to NumPy datatypes. `kwargs` are
+        passed to ``json.load``.
         """
-        with open(fname, "r", encoding=encoding) as f:
-            return cls.from_json(f.read(), **kwargs)
+        with util.xopen(path, "rt", encoding=encoding) as f:
+            return cls.from_json(f.read(), columns, dtypes, **kwargs)
 
     @classmethod
-    def read_pickle(cls, fname):
+    def read_npz(cls, path, allow_pickle=True):
         """
-        Return a new data frame from Pickle file `fname`.
+        Return a new data frame from NumPy file `path`.
+
+        See `numpy.load` for an explanation of `allow_pickle`:
+        https://numpy.org/doc/stable/reference/generated/numpy.load.html
         """
-        with open(fname, "rb") as f:
+        with np.load(path, allow_pickle=allow_pickle) as data:
+            return cls(**data)
+
+    @classmethod
+    def read_pickle(cls, path):
+        """
+        Return a new data frame from Pickle file `path`.
+
+        Will automatically decompress if `path` ends in ``.bz2|.gz|.xz``.
+        """
+        with util.xopen(path, "rb") as f:
             return cls(pickle.load(f))
 
     def _reconcile_column(self, column):
@@ -658,7 +737,7 @@ class DataFrame(dict):
         """
         Return data frame with columns renamed.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.rename(listing_id="id")
         """
         from_to_pairs = {v: k for k, v in to_from_pairs.items()}
@@ -670,7 +749,7 @@ class DataFrame(dict):
         """
         Return randomly chosen `n` rows.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.sample(5)
         """
         if n is None:
@@ -684,7 +763,7 @@ class DataFrame(dict):
         """
         Return data frame, keeping only `colnames`.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.select("id", "hood", "zipcode")
         """
         for colname in colnames:
@@ -700,8 +779,8 @@ class DataFrame(dict):
         `other`.
 
         >>> # All listings that have reviews
-        >>> listings = di.DataFrame.read_csv("data/listings.csv")
-        >>> reviews = di.DataFrame.read_csv("data/listings-reviews.csv")
+        >>> listings = di.read_csv("data/listings.csv")
+        >>> reviews = di.read_csv("data/listings-reviews.csv")
         >>> listings.semi_join(reviews, "id")
         """
         by1, by2 = self._split_by(*by)
@@ -718,7 +797,7 @@ class DataFrame(dict):
         Both `rows` and `cols` should be integer vectors correspoding to the
         indices of the rows or columns to keep.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.slice(rows=[0, 1, 2])
         >>> data.slice(cols=[0, 1, 2])
         >>> data.slice(rows=[0, 1, 2], cols=[0, 1, 2])
@@ -738,7 +817,7 @@ class DataFrame(dict):
         `colname_dir_pairs` defines the sort order by column name with `dir`
         being ``1`` for ascending sort, ``-1`` for descending.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.sort(hood=1, zipcode=1)
         """
         @deco.tuplefy
@@ -750,7 +829,7 @@ class DataFrame(dict):
                 column = self[colname]
                 if dir < 0 and not (column.is_boolean() or column.is_number()):
                     # Use rank for non-numeric types so that we can sort descending.
-                    column = column.rank()
+                    column = column.rank(method="min")
                 yield column if dir >= 0 else -column
         indices = np.lexsort(sort_key())
         for colname, column in self.items():
@@ -765,7 +844,7 @@ class DataFrame(dict):
         """
         Return the last `n` rows.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.tail(5)
         """
         if n is None:
@@ -777,18 +856,18 @@ class DataFrame(dict):
         """
         Return data frame converted to a JSON string.
 
-        `kwargs` are passed to ``json.dumps``.
+        `kwargs` are passed to ``json.dump``.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.to_json()[:100]
         """
         return self.to_list_of_dicts().to_json(**kwargs)
 
     def to_list_of_dicts(self):
         """
-        Return data frame converted to a :class:`dataiter.ListOfDicts`.
+        Return data frame converted to a :class:`.ListOfDicts`.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.to_list_of_dicts()
         """
         from dataiter import ListOfDicts
@@ -802,7 +881,7 @@ class DataFrame(dict):
         """
         Return data frame converted to a ``pandas.DataFrame``.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.to_pandas()
         """
         import pandas as pd
@@ -812,7 +891,7 @@ class DataFrame(dict):
         """
         Return data frame as a string formatted for display.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.to_string()
         """
         if not self: return ""
@@ -854,12 +933,22 @@ class DataFrame(dict):
         """
         Return unique rows by `colnames`.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.unique("hood")
         """
         colnames = colnames or self.colnames
-        by = np.column_stack([self[x].as_bytes() for x in colnames])
-        values, indices = np.unique(by, return_index=True, axis=0)
+        if (len(colnames) == 1 and
+            not self[colnames[0]].is_object()):
+            # Use a single column directly.
+            by = self[colnames[0]]
+        elif (len(set(self[x].dtype for x in colnames)) == 1 and
+              not self[colnames[0]].is_object()):
+            # Stack matching dtypes directly in a new array.
+            by = np.column_stack([self[x] for x in colnames])
+        else:
+            # Use rank for differing dtypes.
+            by = np.column_stack([self[x].rank(method="min") for x in colnames])
+        indices = np.sort(np.unique(by, return_index=True, axis=0)[1])
         for colname, column in self.items():
             yield colname, column[indices].copy()
 
@@ -868,7 +957,7 @@ class DataFrame(dict):
         """
         Return data frame, dropping `colnames`.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.unselect("guests", "sqft", "price")
         """
         for colname in self.colnames:
@@ -880,7 +969,7 @@ class DataFrame(dict):
         """
         Return data frame with columns from `other` added.
 
-        >>> data = di.DataFrame.read_csv("data/listings.csv")
+        >>> data = di.read_csv("data/listings.csv")
         >>> data.update(di.DataFrame(x=1))
         """
         for colname, column in self.items():
@@ -890,27 +979,48 @@ class DataFrame(dict):
             column = self._reconcile_column(column)
             yield colname, column.copy()
 
-    def write_csv(self, fname, encoding="utf_8", header=True, sep=","):
+    def _view_rows(self, rows):
+        # Initialize a blank instance and use base class update
+        # to bypass __init__ and __setitem__ checks for speed.
+        data = self.__class__()
+        dict.update(data, {x: self[x][rows] for x in self})
+        return data
+
+    def write_csv(self, path, encoding="utf-8", header=True, sep=","):
         """
-        Write data frame to CSV file `fname`.
+        Write data frame to CSV file `path`.
+
+        Will automatically compress if `path` ends in ``.bz2|.gz|.xz``.
         """
         pddf = self.to_pandas()
-        util.makedirs_for_file(fname)
-        pddf.to_csv(fname, sep=sep, header=header, index=False, encoding=encoding)
+        util.makedirs_for_file(path)
+        pddf.to_csv(path, sep=sep, header=header, index=False, encoding=encoding)
 
-    def write_json(self, fname, encoding="utf_8", **kwargs):
+    def write_json(self, path, encoding="utf-8", **kwargs):
         """
-        Write data frame to JSON file `fname`.
+        Write data frame to JSON file `path`.
+
+        Will automatically compress if `path` ends in ``.bz2|.gz|.xz``.
 
         `kwargs` are passed to ``json.JSONEncoder``.
         """
-        return self.to_list_of_dicts().write_json(fname, encoding=encoding, **kwargs)
+        return self.to_list_of_dicts().write_json(path, encoding=encoding, **kwargs)
 
-    def write_pickle(self, fname):
+    def write_npz(self, path, compress=False):
         """
-        Write data frame to Pickle file `fname`.
+        Write data frame to NumPy file `path`.
         """
-        util.makedirs_for_file(fname)
-        with open(fname, "wb") as f:
+        util.makedirs_for_file(path)
+        savez = np.savez_compressed if compress else np.savez
+        savez(path, **self)
+
+    def write_pickle(self, path):
+        """
+        Write data frame to Pickle file `path`.
+
+        Will automatically compress if `path` ends in ``.bz2|.gz|.xz``.
+        """
+        util.makedirs_for_file(path)
+        with util.xopen(path, "wb") as f:
             out = {k: np.array(v, v.dtype) for k, v in self.items()}
             pickle.dump(out, f, pickle.HIGHEST_PROTOCOL)
