@@ -21,12 +21,12 @@
 # THE SOFTWARE.
 
 import json
-import warnings
 
 from attd import AttributeDict
 from dataiter import DataFrame
-from dataiter import Vector
+from dataiter import DataFrameColumn
 from dataiter import util
+from dataiter import Vector
 
 
 class GeoJSON(DataFrame):
@@ -35,7 +35,7 @@ class GeoJSON(DataFrame):
     A class for GeoJSON data.
 
     GeoJSON is a simple wrapper class that reads GeoJSON features into a
-    :class:`dataiter.DataFrame`. Any operations on the data are thus done with
+    :class:`.DataFrame`. Any operations on the data are thus done with
     methods provided by the data frame class. Geometry is available in the
     "geometry" column, but no special geometric operations are supported.
 
@@ -66,42 +66,50 @@ class GeoJSON(DataFrame):
     @classmethod
     def _check_raw_data(cls, data):
         if data.type not in cls.TOP_LEVEL_TYPES:
-            raise Exception(f"Top-level type {data.type!r} not supported")
+            raise TypeError(f"Top-level type {data.type!r} not supported")
+        warned_feature_keys = []
         for feature in data.features:
-            cls._check_raw_feature(feature)
+            cls._check_raw_feature(feature, warned_feature_keys)
 
     @classmethod
-    def _check_raw_feature(cls, feature):
+    def _check_raw_feature(cls, feature, warned_feature_keys):
         if feature.type not in cls.FEATURE_TYPES:
-            raise Exception(f"Feature type {feature.type!r} not supported")
+            raise TypeError(f"Feature type {feature.type!r} not supported")
         for key in set(feature) - set(cls.FEATURE_KEYS):
-            warnings.warn(f"Ignoring feature key {key!r}")
+            if key in warned_feature_keys: continue
+            print(f"Warning: Ignoring feature key {key!r}")
+            warned_feature_keys.append(key)
         for key, value in feature.properties.items():
             if isinstance(value, tuple(cls.PROPERTY_TYPES)): continue
-            raise Exception(f"Property type {type(value)} of {key!r} not supported")
+            raise TypeError(f"Property type {type(value)} of {key!r} not supported")
 
     @classmethod
-    def read(cls, fname, encoding="utf_8", **kwargs):
+    def read(cls, path, encoding="utf-8", columns=[], dtypes={}, **kwargs):
         """
-        Return data from GeoJSON file `fname`.
+        Return data from GeoJSON file `path`.
 
-        `kwargs` are passed to ``json.load``.
+        Will automatically decompress if `path` ends in ``.bz2|.gz|.xz``.
 
-        >>> data = di.GeoJSON.read("data/neighbourhoods.geojson")
-        >>> data.head()
+        `columns` is an optional list of columns to limit to. `dtypes` is an
+        optional dict mapping column names to NumPy datatypes. `kwargs` are
+        passed to ``json.load``.
         """
-        with open(fname, "r", encoding=encoding) as f:
+        with util.xopen(path, "rt", encoding=encoding) as f:
             raw = AttributeDict(json.load(f, **kwargs))
         cls._check_raw_data(raw)
         data = {}
         for feature in raw.features:
             for key in feature.properties:
                 data.setdefault(key, [])
+        if columns:
+            data = {k: v for k, v in data.items() if k in columns}
         for feature in raw.features:
             for key in data:
                 value = feature.properties.get(key, None)
                 data[key].append(value)
         data["geometry"] = [x.geometry for x in raw.features]
+        for name, dtype in dtypes.items():
+            data[name] = DataFrameColumn(data[name], dtype)
         data = cls(**data)
         del raw.features
         data.metadata = raw
@@ -112,11 +120,13 @@ class GeoJSON(DataFrame):
         data = self.modify(geometry=Vector.fast(geometry, object))
         return DataFrame.to_string(data, max_rows, max_width)
 
-    def write(self, fname, encoding="utf_8", **kwargs):
+    def write(self, path, encoding="utf-8", **kwargs):
         """
-        Write data to GeoJSON file `fname`.
+        Write data to GeoJSON file `path`.
 
-        `kwargs` are passed to ``json.dumps``.
+        Will automatically compress if `path` ends in ``.bz2|.gz|.xz``.
+
+        `kwargs` are passed to ``json.dump``.
         """
         kwargs.setdefault("default", str)
         kwargs.setdefault("ensure_ascii", False)
@@ -124,10 +134,10 @@ class GeoJSON(DataFrame):
         indent1 = " " * indent_width * 1
         indent2 = " " * indent_width * 2
         if "geometry" not in self:
-            raise Exception("Geometry missing")
+            raise ValueError("Geometry missing")
         data = self.to_list_of_dicts()
-        util.makedirs_for_file(fname)
-        with open(fname, "w", encoding=encoding) as f:
+        util.makedirs_for_file(path)
+        with util.xopen(path, "wt", encoding=encoding) as f:
             f.write("{\n")
             for key, value in self.metadata.items():
                 blob = json.dumps(value, **kwargs)
